{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE RankNTypes #-}

module Shift.Git where

import Control.Lens ((^.))
import Control.Monad (void)
import Control.Monad.Catch (throwM)
import Control.Monad.Reader (runReaderT)
import Control.Monad.State (runStateT)
import Control.Monad.Trans (liftIO)
import Data.ByteString.Char8 (ByteString)
import qualified Data.ByteString.Char8 as BS
import Data.Default (def)
import Data.Either (rights)
import Data.Git
    ( tagList,
      getObject,
      withRepo,
      RefName(refNameRaw),
      Ref,
      Git,
      Commit(commitMessage) )
import Data.Git.Ref (HashAlgorithm, fromHex, isHex)
import Data.Git.Storage.Object ( Object(ObjCommit) )
import Data.List (sortBy)
import Data.Maybe (catMaybes)
import Data.Set as S (toList)
import Data.String.Conversions (cs)
import Data.Text (Text)
import qualified Data.Text as T
import qualified Data.Text.IO as TIO
import Data.Tuple (swap)
import Data.Versions (versioning)
import Data.Void (Void)
import GitHub.Auth (Auth (OAuth))
import Network.HTTP.Client (newManager)
import Network.HTTP.Client.TLS (tlsManagerSettings)
import Shift.CLI
import Shift.Processing ( generateReport, parseCommit )
import Shift.Rendering ( printReport )
import Shift.Types
import Shift.Utilities (orError, pairs)
import System.Process ( readCreateProcess, shell )
import Text.Megaparsec (ParseError)
import Text.Megaparsec.Error (ParseErrorBundle)
import CMarkGFM (NodeType(DOCUMENT), NodeType(DOCUMENT), nodeToCommonmark, nodeToCommonmark, Node(Node))

parseTag :: RefName -> Either (ParseErrorBundle Text Void) TagRef
parseTag ref = case versioning . cs . refNameRaw $ ref of
  Left e -> Left e
  Right v -> Right (TagRef ref v)

tempMain :: ShiftOptions -> IO ()
tempMain opts = do
  node <- renderToNode opts

  liftIO $ TIO.putStr $ nodeToCommonmark [] Nothing node

renderToNode :: ShiftOptions -> IO Node
renderToNode opts = withRepo ".git" $ \repo -> do
  tags <- tagList repo

  let sortedVersions = sortBy (flip compare) . rights $ parseTag <$> toList tags
      pairedTags = swap <$> pairs sortedVersions

  nodes <- case opts ^. soHostingType of
    GitHubType -> do
      state <- initGitHubState

      fst <$> runReaderT
        (runStateT (mapM (renderDiff repo) pairedTags) state)
        opts
    GitType ->
      fst <$> runReaderT
        (runStateT (mapM (renderDiff repo) pairedTags) GitClientState)
        opts

  pure (Node Nothing DOCUMENT (concat nodes))
  where
    initGitHubState = do
      manager <- newManager tlsManagerSettings

      token <- (opts ^. soGitHubToken) `orError` SEMissingGitHubToken
      repositoryOwner <- (opts ^. soGitHubOwner) `orError` SEMissingGitHubOwner
      repositoryName <-
        (opts ^. soGitHubRepository)
          `orError` SEMissingGitHubRepository

      pure
        GitHubClientState
          { _gcsCache = def,
            _gcsAuth = OAuth (cs token),
            _gcsManager = manager,
            _gcsOwner = cs repositoryOwner,
            _gcsRepository = cs repositoryName
          }

renderDiff :: (ClientState s, HashAlgorithm hash) => Git hash -> (TagRef, TagRef) -> GitM s [Node]
renderDiff repo (tx, ty) = do
  diff <- lookupCommitsDiff repo tx ty

  case diff of
    [] -> throwM SEUnableToComputeDiff
    diff_ -> printReport (tx, ty) (generateReport . rights $ parseCommit <$> diff_)

lookupCommitsDiff :: (HashAlgorithm hash) => Git hash -> TagRef -> TagRef -> GitM s [(Ref hash, Commit hash)]
lookupCommitsDiff repo x y = do
  rawOutput <- liftIO $ readCreateProcess (shell gitCommand) ""

  catMaybes
    <$> mapM
      (lookupRawRef repo)
      (filter (not . T.null) . T.splitOn "\n" . cs $ rawOutput)
  where
    gitCommand =
      mconcat
        [ "git rev-list ",
          refNameRaw . _tRef $ x,
          "...",
          refNameRaw . _tRef $ y
        ]

lookupRawRef :: (HashAlgorithm hash) => Git hash -> Text -> GitM s (Maybe (Ref hash, Commit hash))
lookupRawRef repo rr =
  if isHex . cs $ rr
    then do
      let ref = fromHex . cs $ rr

      object <- liftIO $ getObject repo ref True

      case object of
        Just (ObjCommit commit) -> pure $ Just (ref, commit)
        _ -> pure Nothing
    else pure Nothing

commitSummary :: Commit hash -> ByteString
commitSummary = head . BS.split '\n' . commitMessage
